@@ -1,13 +1,14 @@
 import base64
 import binascii
 import datetime
-import decimal
+import io
 import json
 import logging
 import os
 import random
 import sys
 import time
+import re
 from io import IOBase, BytesIO
 from urllib.parse import urljoin
 
@@ -16,13 +17,29 @@ import requests
 
 from .entity.exception import BoonSdkException
 from .entity.storage import StoredFile
+from .util import to_json
 
 logger = logging.getLogger(__name__)
 
 DEFAULT_SERVER = 'https://api.boonai.app'
 
+__all__ = [
+    'BoonClient',
+    'FileInputStream',
+    'to_json',
+    'BoonSdkConnectionException',
+    'BoonSdkException',
+    'BoonSdkRequestException',
+    'BoonSdkWriteException',
+    'BoonSdkSecurityException',
+    'BoonSdkInvalidRequestException',
+    'BoonSdkDuplicateException',
+    'BoonSdkNotFoundException',
+    'BoonClientException'
+]
 
-class BoonClient(object):
+
+class BoonClient:
     """
     BoonClient is used to communicate to a Boon AI API server.
     """
@@ -43,6 +60,7 @@ class BoonClient(object):
         self.project_id = kwargs.get('project_id', os.environ.get("BOONAI_PROJECT"))
         self.max_retries = kwargs.get('max_retries', 3)
         self.verify = True
+        self.token_override = None
 
     def stream(self, url, dst):
         """
@@ -58,8 +76,7 @@ class BoonClient(object):
                                         headers=self.headers(), stream=True)
 
                 if not response.ok:
-                    raise BoonClientException(
-                        "Failed to stream asset: %s, %s" % (url, response))
+                    raise BoonClientException("Failed to stream asset: %s, %s" % (url, response))
 
                 for block in response.iter_content(1024):
                     handle.write(block)
@@ -107,6 +124,26 @@ class BoonClient(object):
                 self.get_url(path), headers=self.headers(content_type=""),
                 data=f), True)
 
+    def send_data(self, path, data, size=None):
+        """
+        Send a BytesIO or StringIO to the given URI.  Optionally provide
+        the size which is passed on via the content-length header, otheriwse
+        the size will be auto-detected.
+
+        Args:
+            path (path): The URI fragment for the request.
+            data (io.BytesIO): The bytes to send.s
+
+        Returns:
+            dict: A dictionary which can be used to fetch the file.
+        """
+        reader = InputReader(data)
+        headers = self.headers(content_type="")
+        headers['Content-Length'] = str(size or sys.getsizeof(data))
+
+        return self.__handle_rsp(requests.post(
+            self.get_url(path), headers=headers, data=reader.read()), True)
+
     def upload_file(self, path, file, body={}, json_rsp=True):
         """
         Upload a single file and a request to the given endpoint path.
@@ -121,7 +158,7 @@ class BoonClient(object):
             dict: The response body of the request.
         """
         try:
-            post_files = [("file", (os.path.basename(file), open(file, 'rb')))]
+            post_files = [("file", (os.path.basename(file), FileInputStream(file, 'rb')))]
             if body is not None:
                 post_files.append(
                     ["body", (None, to_json(body), 'application/json')])
@@ -135,7 +172,7 @@ class BoonClient(object):
 
     def upload_files(self, path, files, body, json_rsp=True):
         """
-        Upload an array of files and a reques to the given endpoint path.
+        Upload an array of files and a request to the given endpoint path.
 
         Args:
             path (str): The URL to upload to
@@ -154,7 +191,7 @@ class BoonClient(object):
                         ("files", (os.path.basename(f.name), f)))
                 else:
                     post_files.append(
-                        ("files", (os.path.basename(f), open(f, 'rb'))))
+                        ("files", (os.path.basename(f), FileInputStream(f, 'rb'))))
 
             if body is not None:
                 post_files.append(
@@ -276,6 +313,25 @@ class BoonClient(object):
          """
         return self._make_request('delete', path, body, is_json)
 
+    def patch(self, path, body=None, is_json=True):
+        """
+         Performs a patch request.
+
+         Args:
+             path (str): An archivist URI path.
+             body (object): The request body which will be serialized to json.
+             is_json (bool): Set to true to specify a JSON return value
+
+         Returns:
+             object: The http response object or an object deserialized from
+             the response json if the ``json`` argument is true.
+
+         Raises:
+             Exception: An error occurred making the request or parsing the
+                JSON response
+         """
+        return self._make_request('patch', path, body, is_json)
+
     def iter_paged_results(self, url, req, limit, cls):
         """
         Handles paging through the results of the standard _search
@@ -374,11 +430,21 @@ class BoonClient(object):
         """
         Returns the full URL including the configured server part.
         """
-        url = urljoin(self.server, path)
+        url = urljoin(self.get_server(), path)
         if logger.getEffectiveLevel() == logging.DEBUG:
             logger.debug("url: '%s' path: '%s' body: '%s'" % (url, path, body))
         return url
 
+    def get_server(self):
+        """
+        Return the server address.  This is here to be overriden by subclasses.  The
+        public 'server' propertly has been left in for backwards compat for now.
+
+        Returns:
+            str: The server URL
+        """
+        return self.server
+
     def headers(self, content_type="application/json"):
         """
         Generate the return some request headers.
@@ -391,7 +457,7 @@ class BoonClient(object):
             dict: An http header struct.
 
         """
-        header = {'Authorization': "Bearer {}".format(self.__sign_request())}
+        header = {'Authorization': self.sign_request()}
 
         if content_type:
             header['Content-Type'] = content_type
@@ -417,25 +483,25 @@ class BoonClient(object):
 
         return key_data
 
-    def __sign_request(self):
+    def sign_request(self):
         if not self.apikey:
-            raise RuntimeError("Unable to make request, no ApiKey has been specified.")
+            raise RuntimeError('Unable to make request, no ApiKey has been specified.')
         claims = {
-            'aud': self.server,
+            'aud': self.get_server(),
             'exp': datetime.datetime.utcnow() + datetime.timedelta(seconds=60),
-            'accessKey': self.apikey["accessKey"],
+            'accessKey': self.apikey['accessKey'],
         }
 
-        if os.environ.get("BOONAI_TASK_ID"):
-            claims['taskId'] = os.environ.get("BOONAI_TASK_ID")
-            claims['jobId'] = os.environ.get("BOONAI_JOB_ID")
+        if os.environ.get('BOONAI_TASK_ID'):
+            claims['taskId'] = os.environ.get('BOONAI_TASK_ID')
+            claims['jobId'] = os.environ.get('BOONAI_JOB_ID')
 
         if self.project_id:
-            claims["projectId"] = self.project_id
-        return jwt.encode(claims, self.apikey['secretKey'], algorithm='HS512')
+            claims['projectId'] = self.project_id
+        return "Bearer {}".format(jwt.encode(claims, self.apikey['secretKey'], algorithm='HS512'))
 
 
-class SearchResult(object):
+class SearchResult:
     """
     A utility class for wrapping various search result formats
     that come back from the Boon AI servers.
@@ -464,47 +530,6 @@ class SearchResult(object):
         return self.items[idx]
 
 
-def to_json(obj, indent=None):
-    """
-    Convert the given object to a JSON string using
-    the BoonSdkJsonEncoder.
-
-    Args:
-        obj (mixed): any json serializable python object.
-        indent (int): The indentation level for the json, or None for compact.
-    Returns:
-        str: The serialized object
-
-    """
-    val = json.dumps(obj, cls=BoonSdkJsonEncoder, indent=indent)
-    if logger.getEffectiveLevel() == logging.DEBUG:
-        logger.debug("json: %s" % val)
-    return val
-
-
-class BoonSdkJsonEncoder(json.JSONEncoder):
-    """
-    JSON encoder for with Boon AI specific serialization defaults.
-    """
-
-    def default(self, obj):
-        if hasattr(obj, 'for_json'):
-            return obj.for_json()
-        elif isinstance(obj, (set, frozenset)):
-            return list(obj)
-        elif isinstance(obj, datetime.datetime):
-            return obj.isoformat()
-        elif isinstance(obj, datetime.date):
-            return obj.isoformat()
-        elif isinstance(obj, datetime.time):
-            return obj.isoformat()
-        elif isinstance(obj, decimal.Decimal):
-            return float(obj)
-
-        # Let the base class default method raise the TypeError
-        return json.JSONEncoder.default(self, obj)
-
-
 class BoonClientException(BoonSdkException):
     """The base exception class for all BoonClient related Exceptions."""
     pass
@@ -536,7 +561,39 @@ class BoonSdkRequestException(BoonClientException):
         return self.__data["status"]
 
     def __str__(self):
-        return "<BoonSdkRequestException msg=%s>" % self.__data["message"]
+        return "<BoonSdkRequestException msg=%s>" % self.__data.get("message", "Unknown message")
+
+
+class FileInputStream:
+    """
+    A partially implemented File object which just supports reading the
+    entire file and then closing the file handle.
+    """
+    def __init__(self, filename, mode='rb'):
+        self.filename = filename
+        self.mode = mode
+
+    def read(self):
+        with open(self.filename, self.mode) as f:
+            return f.read()
+
+
+class InputReader:
+    """
+    Reads input from a file path, url, or file handle.
+    """
+    def __init__(self, obj):
+        self.obj = obj
+
+    def read(self):
+        if isinstance(self.obj, str):
+            if re.match('http[s]?://', self.obj):
+                with requests.get(self.obj, stream=True) as rsp:
+                    return io.BytesIO(rsp.content).read()
+            else:
+                return FileInputStream(self.obj).read()
+        else:
+            return self.obj.read()
 
 
 class BoonSdkConnectionException(BoonClientException):
